/**
 *   This program, AMController libraries, example sketches (“The Software”) and the related documentation (“The Documentation”) are supplied to you
 *   by the Author in consideration of your agreement to the following terms, and your use or installation of The Software and the use of The Documentation
 *   constitutes acceptance of these terms.
 *   If you do not agree with these terms, please do not use or install The Software.
 *   The Author grants you a personal, non-exclusive license, under author's copyrights in this original software, to use The Software.
 *   Except as expressly stated in this notice, no other rights or licenses, express or implied, are granted by the Author, including but not limited to any
 *   patent rights that may be infringed by your derivative works or by other works in which The Software may be incorporated.
 *   The Software and the Documentation are provided by the Author on an "AS IS" basis.  THE AUTHOR MAKES NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT
 *   LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND OPERATION
 *   ALONE OR IN COMBINATION WITH YOUR PRODUCTS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 *   REPRODUCTION AND MODIFICATION OF THE SOFTWARE AND OR OF THE DOCUMENTATION, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE),
 *   STRICT LIABILITY OR OTHERWISE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

/**
 * Version: 1.0
 *
 * For more information write to fabboco@gmail.com
 **/

#ifndef MEASUREMENT_H
#define MEASUREMENT_H

#include <math.h>
#include <stdlib.h>
#include "hardware/adc.h"

#define VREF 3.3

#define MEAN_SAMPLES 100 // Number of samples to calculate the mean voltage or current for calibration

#define RMS_SAMPLES 100 // Number of samples to calculate the voltage's or current's RMS value

#define MEASUREMENT_SAMPLES 40 // Number of samples to get a single voltage or current measurement

#define ZMPT101B_SCALE 863.47 // Scale to convert the ZMPT101B output to actual voltage
#define SCT_013_30_SCALE 30.0 // Scale to convert the SCT-013-30 output to actual current

#define sleep_us_50hz 2000 // 50Hz sine wave (250 Hz sampling rate)
#define sleep_us_60hz 1667 // 60Hz sine wave (300 Hz sampling rate)

#define sleep sleep_us_50hz

struct voltage_calibration_data
{
    float value;
    float delta;
};

typedef struct voltage_calibration_data voltage_calibration_data;

// Comparison function for float values
int compareFloats(const void *a, const void *b)
{
    // Cast the void pointers to float pointers
    float fa = *(const float *)a;
    float fb = *(const float *)b;

    // Return negative if fa < fb, 0 if equal, positive if fa > fb
    if (fa < fb)
        return -1;
    if (fa > fb)
        return 1;
    return 0;
}

class Measurement
{
private:
    uint voltage_gpio;
    uint current_gpio;
    float v_dc;
    float i_dc;

public:
    Measurement(uint voltage_gpio, uint current_gpio);
    ~Measurement();

    // Sets the stored voltage dc offset
    void setVdc(float v_dc);

    // Sets the stored current dc offset
    void setIdc(float i_dc);

    // Returns the instantaneous voltage value
    float voltage();

    // Returns the instantaneous current value
    float current();

    // Returns the average voltage value assuming it is not varying over time
    // Used to measure the dc offset
    float mean_voltage();

    // Returns the average current value assuming it is not varying over time
    // Used to measure the dc offset
    float mean_current();

    // Returns the RMS voltage value sampling it at a frequency double of its natural frequency (50 or 60Hz)
    float voltage_rms();

    // Returns the RMS voltage value sampling it at a frequency double of its natural frequency (50 or 60Hz)
    float current_rms();

    voltage_calibration_data voltage_calibration(float v_dc);
};

Measurement::Measurement(uint voltage_gpio, uint current_gpio)
{
    this->voltage_gpio = voltage_gpio;
    this->current_gpio = current_gpio;
    this->v_dc = 0;
    this->i_dc = 0;
}

Measurement::~Measurement()
{
}

void Measurement::setVdc(float v_dc)
{
    this->v_dc = v_dc;
}
void Measurement::setIdc(float i_dc)
{
    this->i_dc = i_dc;
}

float Measurement::voltage()
{
    adc_select_input(voltage_gpio - 26);

    float samples[MEASUREMENT_SAMPLES];
    for (int i = 0; i < MEASUREMENT_SAMPLES; i++)
    {
        samples[i] = adc_read();
    }

    // Sort the array using std::qsort
    std::qsort(samples, MEASUREMENT_SAMPLES, sizeof(float), compareFloats);

    float sum = 0;
    // First and last 3 samples removed
    for (int i = 3; i < MEASUREMENT_SAMPLES - 3; i++)
    {
        sum += VREF * samples[i] / 4095;
    }

    return sum / (MEASUREMENT_SAMPLES - 6);
}

float Measurement::current()
{
    adc_select_input(current_gpio - 26);

    float samples[MEASUREMENT_SAMPLES];
    for (int i = 0; i < MEASUREMENT_SAMPLES; i++)
    {
        samples[i] = adc_read();
    }

    // Sort the array using std::qsort
    std::qsort(samples, MEASUREMENT_SAMPLES, sizeof(float), compareFloats);

    float sum = 0;
    // First and last 3 samples removed
    for (int i = 3; i < MEASUREMENT_SAMPLES - 3; i++)
    {
        sum += VREF * samples[i] / 4095.0;
    }

    return sum / (MEASUREMENT_SAMPLES - 6);
}

float Measurement::mean_voltage()
{
    float sum = 0;
    for (int i = 0; i < MEAN_SAMPLES; i++)
    {
        sum += voltage();
    }

    return sum / MEAN_SAMPLES;
}

float Measurement::mean_current()
{
    float sum = 0;
    for (int i = 0; i < MEAN_SAMPLES; i++)
    {
        sum += current();
    }

    return sum / MEAN_SAMPLES;
}

float Measurement::voltage_rms()
{
    float squared_sum = 0;

    for (int i = 0; i < RMS_SAMPLES; i++)
    {
        float v = voltage() - v_dc; // DC offset removed
        squared_sum += v * v;       // Sum the squared voltages
        sleep_us(sleep);            // Delay between samples
    }

    // Calculate the mean of the squared values and return the square root (RMS)
    float mean_squared = squared_sum / RMS_SAMPLES;
    float voltage_rms = sqrtf(mean_squared);

    // Scale the sensor's output to the actual voltage
    voltage_rms = ZMPT101B_SCALE * sqrtf(mean_squared);

    // Any RMS value under 1V is assumed 0
    if (voltage_rms < 1)
    {
        voltage_rms = 0.0;
    }

    return voltage_rms;
}

voltage_calibration_data Measurement::voltage_calibration(float v_dc)
{
    float p_squared_sum = 0;
    float n_squared_sum = 0;

    for (int i = 0; i < 4 * RMS_SAMPLES; i++)
    {
        float v = voltage() - v_dc;

        if (v > 0)
        {
            p_squared_sum += v * v;
        }
        else
        {
            n_squared_sum += v * v;
        }

        sleep_us(sleep); // Delay between samples
    }

    // Calculate the mean of the squared values and return the square root (RMS)
    float p_mean_squared = p_squared_sum / (4 * RMS_SAMPLES);
    float p_voltage_rms = sqrtf(p_mean_squared);

    float n_mean_squared = n_squared_sum / (4 * RMS_SAMPLES);
    float n_voltage_rms = sqrtf(n_mean_squared);

    voltage_calibration_data ret_data;

    ret_data.value = p_voltage_rms;
    ret_data.delta = p_voltage_rms - n_voltage_rms;

    return ret_data;
}

float Measurement::current_rms()
{
    float squared_sum = 0;

    for (int i = 0; i < RMS_SAMPLES; i++)
    {
        float v = current() - i_dc;
        squared_sum += v * v; // Sum the squared voltages
        sleep_us(sleep);      // Delay between samples
    }

    // Calculate the mean of the squared values and return the square root (RMS)
    float mean_squared = squared_sum / RMS_SAMPLES;
    float voltage_rms = sqrtf(mean_squared);

    // Convert the measured voltage in current. The current values is SCT_013_30_SCALE for each volt
    float current_rms = SCT_013_30_SCALE * voltage_rms;

    // Any RMS value under 0.007 is assumed 0
    if (current_rms < 0.007)
    {
        current_rms = 0.0;
    }

    return current_rms;
}

#endif